'''
Created on Sep 9, 2009

@author: tw55413
'''
import logging
logger = logging.getLogger('camelot.view.model_thread.signal_slot_model_thread')

from PyQt4 import QtCore

from camelot.view.model_thread import AbstractModelThread, gui_function, model_function, setup_model

class Task(QtCore.QObject):

    finished = QtCore.SIGNAL('finished')
    exception = QtCore.SIGNAL('exception')

    def __init__(self, request, name=''):
        QtCore.QObject.__init__(self)
        self._request = request
        self._name = name

    def clear(self):
        """clear this tasks references to other objects"""
        self._request = None
        self._name = None

    def execute(self):
        logger.debug('executing %s' % (self._name))
        try:
            result = self._request()
            #import sip
            #print '===', self._name, '==='
            #print self.receivers(QtCore.SIGNAL('finished'))
            #sip.dump(self)
            self.emit(QtCore.SIGNAL('finished'), result )
        except Exception, e:
            logger.error( 'exception caught in model thread', exc_info = e )
            exception_info = (e, '')
            self.emit(QtCore.SIGNAL('exception'), exception_info)

def synchronized( original_function ):
    """Decorator for synchronized access to an object, the object should
    have an attribute _mutex which is of type QMutex
    """

    from functools import wraps

    @wraps( original_function )
    def wrapper(self, *args, **kwargs):
        QtCore.QMutexLocker(self._mutex)
        return original_function(self, *args, **kwargs)

    return wrapper

class TaskHandler(QtCore.QObject):

    def __init__(self, queue):
        """:param queue: the queue from which to pop a task when handle_task
        is called"""
        QtCore.QObject.__init__(self)
        self._mutex = QtCore.QMutex()
        self._queue = queue
        self._tasks_done = []
        self._busy = False
        #logger.debug("TaskHandler created.")

    @synchronized
    def busy(self):
        """:return True/False: indicating if this task handler is busy"""
        return self._busy

    @synchronized
    def handle_task(self):
        self._busy = True
        self.emit( AbstractModelThread.thread_busy_signal, True  )
        task = self._queue.pop()
        while task:
            task.execute()
            # we keep track of the tasks done to prevent them being garbage collected
            # apparently when they are garbage collected, they are recycled, but their
            # signal slot connections seem to survive this recycling.
            # @todo: this should be investigated in more detail, since we are causing
            #        a deliberate memory leak here
            task.clear()
            self._tasks_done.append(task)
            task = self._queue.pop()
        self.emit( AbstractModelThread.thread_busy_signal, False  )
        self._busy = False

class SignalSlotModelThread( QtCore.QThread, AbstractModelThread ):
    """A model thread implementation that uses signals and slots
    to communicate between the model thread and the gui thread
    """

    task_available = QtCore.SIGNAL('task_available')

    def __init__( self, setup_thread = setup_model ):
        """
        @param setup_thread: function to be called at startup of the thread to initialize
        everything, by default this will setup the model.  set to None if nothing should
        be done.
        """
        QtCore.QThread.__init__( self )
        AbstractModelThread.__init__( self, setup_thread )
        self._task_handler = None
        self._mutex = QtCore.QMutex()
        self._request_queue = []
        self._connected = False

    def run( self ):
        self.logger.debug( 'model thread started' )
        self._task_handler = TaskHandler(self)
        self.connect(self._task_handler, self.thread_busy_signal, self._thread_busy, QtCore.Qt.QueuedConnection)
        self._thread_busy(True)
        try:
            self._setup_thread()
        except Exception, e:
            self.logger.error('thread setup incomplete', exc_info=e)
        self._thread_busy(False)
        self.logger.debug('thread setup finished')
        # Some tasks might have been posted before the signals were connected to the task handler,
        # so once force the handling of tasks
        self._task_handler.handle_task()
        self.exec_()
        self.logger.debug('model thread stopped')

    def _thread_busy(self, busy_state):
        self.emit(self.thread_busy_signal, busy_state)
                
    @synchronized
    def post( self, request, response = None, exception = None ):
        if not self._connected and self._task_handler:
            # creating this connection in the model thread throws QT exceptions
            self.connect(self, self.task_available, self._task_handler.handle_task, QtCore.Qt.QueuedConnection)
            self._connected = True
        # response should be a slot method of a QObject
        if response:
            name = '%s -> %s.%s'%(request.__name__, response.im_self.__class__.__name__, response.__name__)
        else:
            name = request.__name__
        task = Task(request, name=name)
        # QObject::connect is a thread safe function
        if response:
            assert response.im_self
            assert isinstance(response.im_self, QtCore.QObject)
            self.connect(task, QtCore.SIGNAL('finished'), response, QtCore.Qt.QueuedConnection)
        if exception:
            self.connect(task, QtCore.SIGNAL('exception'), exception, QtCore.Qt.QueuedConnection)
        task.moveToThread(self)
        # only put the task in the queue when it is completely set up 
        self._request_queue.append(task)
        #print 'task created --->', id(task)
        self.emit(self.task_available)

    @synchronized
    @model_function
    def pop( self ):
        """Pop a task from the queue, return None if the queue is empty"""
        if len(self._request_queue):
            task = self._request_queue.pop(0)
            return task

    @synchronized
    @gui_function
    def busy( self ):
        """Return True or False indicating wether either the model or the
        gui thread is doing something"""
        while not self._task_handler:
            import time
            time.sleep(1)
        app = QtCore.QCoreApplication.instance()
        return app.hasPendingEvents() or len(self._request_queue) or self._task_handler.busy()

    @gui_function
    def wait_on_work(self):
        """Wait for all work to be finished, this function should only be used
        to do unit testing and such, since it will block the calling thread until
        all work is done"""
        app = QtCore.QCoreApplication.instance()
        while self.busy():
            app.processEvents()

#    app = QCoreApplication.instance()
#    waiting = True
#    while waiting:
#      waiting = False
#      if app.hasPendingEvents():
#        app.processEvents()
#        waiting = True




from customdelegate import CustomDelegate, DocumentationMetaclass, ValueLoading
from camelot.view.controls import editors
from camelot.core.utils import variant_to_pyobject

from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import Qt

class DateTimeDelegate(CustomDelegate):
  
    __metaclass__ = DocumentationMetaclass
    
    editor = editors.DateTimeEditor
    
    def __init__(self, parent, editable, **kwargs):
        CustomDelegate.__init__(self, parent, editable=editable, **kwargs)
        locale = QtCore.QLocale()
        self.dateTime_format = locale.dateTimeFormat(locale.ShortFormat)
        
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        dateTime = variant_to_pyobject(index.model().data(index, Qt.EditRole))
        if dateTime not in (ValueLoading, None):
            dateTime = QtCore.QDateTime(dateTime.year, dateTime.month, dateTime.day,
                                        dateTime.hour, dateTime.minute, dateTime.second)
            formattedDateTime = dateTime.toString(self.dateTime_format)
        else:
            formattedDateTime = ''
        background_color = QtGui.QColor(index.model().data(index, Qt.BackgroundRole))
        rect = option.rect
        rect = QtCore.QRect(rect.left()+3, rect.top()+6, 16, 16)
        
        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect(option.rect, option.palette.highlight())
            fontColor = QtGui.QColor()
            if self.editable:
                Color = option.palette.highlightedText().color()
                fontColor.setRgb(Color.red(), Color.green(), Color.blue())
            else:
                fontColor.setRgb(130,130,130)
        else:
            if self.editable:
                painter.fillRect(option.rect, background_color)
                fontColor = QtGui.QColor()
                fontColor.setRgb(0,0,0)
            else:
                painter.fillRect(option.rect, option.palette.window())
                fontColor = QtGui.QColor()
                fontColor.setRgb(130,130,130)
              
              
        painter.setPen(fontColor.toRgb())
        rect = QtCore.QRect(option.rect.left(),
                            option.rect.top(),
                            option.rect.width(),
                            option.rect.height())
        
        painter.drawText(rect.x()+2,
                         rect.y(),
                         rect.width()-4,
                         rect.height(),
                         Qt.AlignVCenter | Qt.AlignRight,
                         str(formattedDateTime))
        painter.restore()
        
        
        
    



import datetime

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.core.utils import create_constant_function

class TimeDelegate(CustomDelegate):
 
    __metaclass__ = DocumentationMetaclass
   
    editor = editors.TimeEditor
      
    def __init__(self, parent=None, editable=True, **kwargs):
        CustomDelegate.__init__(self, parent, editable)
        locale = QtCore.QLocale()
        self.time_format = locale.timeFormat(locale.ShortFormat)
        
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        
        formattedTime = unicode(index.model().data(index, Qt.EditRole).toTime().toString(self.time_format))
        
        background_color = QtGui.QColor(index.model().data(index, Qt.BackgroundRole))
        
        rect = option.rect
        rect = QtCore.QRect(rect.left()+3, rect.top()+6, 16, 16)
        
        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect(option.rect, option.palette.highlight())
            fontColor = QtGui.QColor()
            if self.editable:
                Color = option.palette.highlightedText().color()
                fontColor.setRgb(Color.red(), Color.green(), Color.blue())
            else:
                fontColor.setRgb(130,130,130)
        else:
            if self.editable:
                painter.fillRect(option.rect, background_color)
                fontColor = QtGui.QColor()
                fontColor.setRgb(0,0,0)
            else:
                painter.fillRect(option.rect, option.palette.window())
                fontColor = QtGui.QColor()
                fontColor.setRgb(130,130,130)
              
              
        painter.setPen(fontColor.toRgb())
        rect = QtCore.QRect(option.rect.left()+23,
                            option.rect.top(),
                            option.rect.width()-23,
                            option.rect.height())
        painter.drawText(rect.x()+2,
                         rect.y(),
                         rect.width()-4,
                         rect.height(),
                         Qt.AlignVCenter | Qt.AlignRight,
                         formattedTime)
        painter.restore()
      
      
      
    def setModelData(self, editor, model, index):
        value = editor.time()
        t = datetime.time(hour=value.hour(),
                          minute=value.minute(),
                          second=value.second())
        model.setData(index, create_constant_function(t))



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import logging
logger = logging.getLogger('camelot.view.elixir_admin')

import sqlalchemy.sql.expression

from camelot.admin.object_admin import ObjectAdmin
from camelot.view.model_thread import post, model_function, gui_function
from camelot.core.utils import ugettext as _
from camelot.admin.validator.entity_validator import EntityValidator


class EntityAdmin(ObjectAdmin):
    """Admin class specific for classes that are mapped by sqlalchemy.
    This allows for much more introspection than the standard ObjectAdmin.
    """

    validator = EntityValidator

    @model_function
    def get_query(self):
        """:return: an sqlalchemy query for all the objects that should be
        displayed in the table or the selection view.  Overwrite this method to
        change the default query, which selects all rows in the database.
        """
        return self.entity.query

    @model_function
    def get_subclass_entity_admin(self, entity):
        """Get the admin class for an entity that is a subclass of this admin's
        entity or this admin's entity itself.
        """
        for subclass_admin in self.get_subclasses():
            if subclass_admin.entity == entity:
                return subclass_admin
        return self

    @model_function
    def get_subclasses(self):
        """Returns admin objects for the subclasses of the Entity represented
        by this admin object.
        """
        if not self._subclasses:
            from elixir import entities
            self._subclasses = [
                e.Admin(self.app_admin, e)
                for e in entities
                    if (
                        issubclass(e, (self.entity,))
                        and hasattr(e, 'Admin')
                        and e != self.entity
                    )
            ]
        return self._subclasses

    @model_function
    def get_verbose_identifier(self, obj):
        if hasattr(obj, 'id') and obj.id:
            if hasattr(obj, '__unicode__'):
                return u'%s %s : %s' % (
                    unicode(self.get_verbose_name().capitalize()),
                    unicode(obj.id),
                    unicode(obj)
                )
            else:
                return u'%s %s' % (
                    self.get_verbose_name().capitalize(),
                    unicode(obj.id)
                )
        else:
            return self.get_verbose_name()

    @model_function
    def get_field_attributes(self, field_name):
        """Get the attributes needed to visualize the field field_name
        :param field_name: the name of the field

        :return: a dictionary of attributes needed to visualize the field,
        those attributes can be:
         * python_type : the corresponding python type of the object
         * editable : bool specifying wether the user can edit this field
         * widget : which widget to be used to render the field
         * ...
        """

        try:
            return self._field_attributes[field_name]
        except KeyError:
            from camelot.view.controls import delegates
            #
            # Default attributes for all fields
            #
            attributes = dict(
                python_type = str,
                length = None,
                tooltip = None,
                background_color = None,
                minimal_column_width = 0,
                editable = False,
                nullable = True,
                widget = 'str',
                blank = True,
                delegate = delegates.PlainTextDelegate,
                validator_list = [],
                name = unicode(_(field_name.replace('_', ' '))).capitalize()
            )

            #
            # Field attributes forced by the field_attributes property
            #
            forced_attributes = {}
            try:
                forced_attributes = self.field_attributes[field_name]
            except KeyError:
                pass

            def get_entity_admin(target):
                """Helper function that instantiated an Admin object for a
                target entity class.

                :param target: an entity class for which an Admin object is
                needed.
                """
                try:
                    fa = self.field_attributes[field_name]
                    target = fa.get('target', target)
                    admin_class = fa['admin']
                    return admin_class(self.app_admin, target)
                except KeyError:
                    return self.get_related_entity_admin(target)
            #
            # Get the default field_attributes trough introspection if the
            # field is a mapped field
            #
            from sqlalchemy import orm
            from sqlalchemy.exceptions import InvalidRequestError
            from field_attributes import _sqlalchemy_to_python_type_
            mapper = orm.class_mapper(self.entity)
            try:
                property = mapper.get_property(
                    field_name,
                    resolve_synonyms=True
                )
                if isinstance(property, orm.properties.ColumnProperty):
                    type = property.columns[0].type
                    python_type = _sqlalchemy_to_python_type_.get(
                        type.__class__,
                        None
                    )
                    if python_type:
                        attributes.update(python_type(type))
                    if not isinstance(
                        property.columns[0],
                        sqlalchemy.sql.expression._Label
                    ):
                        attributes['nullable'] = property.columns[0].nullable
                        attributes['default'] = property.columns[0].default
                elif isinstance(property, orm.properties.PropertyLoader):
                    target = property._get_target().class_
                    foreign_keys = list(property._foreign_keys)
                    if property.direction == orm.interfaces.ONETOMANY:
                        attributes.update(
                            python_type = list,
                            editable = True,
                            nullable = True,
                            delegate = delegates.One2ManyDelegate,
                            target = target,
                            create_inline = False,
                            backref = property.backref.key,
                            direction = property.direction,
                            admin = get_entity_admin(target)
                        )
                    elif property.direction == orm.interfaces.MANYTOONE:
                        attributes.update(
                            python_type = str,
                            editable = True,
                            delegate = delegates.Many2OneDelegate,
                            target = target,
                            #
                            #@todo: take into account all foreign keys instead
                            # of only the first one
                            #
                            nullable = foreign_keys[0].nullable,
                            direction = property.direction,
                            admin = get_entity_admin(target)
                        )
                    elif property.direction == orm.interfaces.MANYTOMANY:
                        attributes.update(
                            python_type = list,
                            editable = True,
                            target = target,
                            nullable = True,
                            create_inline = False,
                            direction = property.direction,
                            delegate = delegates.ManyToManyDelegate,
                            admin = get_entity_admin(target)
                        )
                    else:
                        raise Exception('PropertyLoader has unknown direction')
            except InvalidRequestError:
                #
                # If the field name is not a property of the mapper, then use
                # the default stuff
                #
                pass

            if 'choices' in forced_attributes:
                attributes['delegate'] = delegates.ComboBoxDelegate
                attributes['editable'] = True

            #
            # Overrule introspected field_attributes with those defined
            #
            attributes.update(forced_attributes)

            #
            # In case of a 'target' field attribute, instantiate an appropriate
            # 'admin' attribute
            #
            if 'target' in attributes:
                attributes['admin'] = get_entity_admin(attributes['target'])

            # if name should be translated, do so now
            attributes['name'] = unicode(attributes['name']).capitalize()

            self._field_attributes[field_name] = attributes
            return attributes

    @model_function
    def get_list_charts(self):
        return self.list_charts

    @model_function
    def get_filters(self):
        """Returns the filters applicable for these entities each filter is

        :return: [(filter_name, [(option_name, query_decorator), ...), ... ]
        """
        from filters import structure_to_filter

        def filter_generator():
            for structure in self.list_filter:
                filter = structure_to_filter(structure)
                yield (filter, filter.get_name_and_options(self))

        return list(filter_generator())

    @model_function
    def set_defaults(self, entity_instance):
        """Set the defaults of an object"""
        from sqlalchemy.schema import ColumnDefault
        for field, attributes in self.get_fields():
            try:
                default = attributes['default']
                if isinstance(default, ColumnDefault):
                    default_value = default.execute()
                elif callable(default):
                    import inspect
                    args, _varargs, _kwargs, _defs = \
                        inspect.getargspec(default)
                    if len(args):
                        default_value = default(entity_instance)
                    else:
                        default_value = default()
                else:
                    default_value = default
                logger.debug(
                    'set default for %s to %s' % (
                        field,
                        unicode(default_value)
                    )
                )
                try:
                    setattr(entity_instance, field, default_value)
                except AttributeError, e:
                    logger.error(
                        'Programming Error : could not set'
                        ' attribute %s to %s on %s' % (
                            field,
                            default_value,
                            entity_instance.__class__.__name__
                        ),
                        exc_info=e
                    )
            except KeyError, _e:
                pass

    @gui_function
    def create_new_view(admin, parent=None, oncreate=None, onexpunge=None):
        """Create a Qt widget containing a form to create a new instance of the
        entity related to this admin class

        The returned class has an 'entity_created_signal' that will be fired
        when a valid new entity was created by the form
        """
        from PyQt4 import QtCore
        from PyQt4 import QtGui
        from PyQt4.QtCore import SIGNAL
        from camelot.view.controls.view import AbstractView
        from proxy.collection_proxy import CollectionProxy
        new_object = []

        @model_function
        def collection_getter():
            if not new_object:
                entity_instance = admin.entity()
                if oncreate:
                    oncreate(entity_instance)
                # Give the default fields their value
                admin.set_defaults(entity_instance)
                new_object.append(entity_instance)
            return new_object

        model = CollectionProxy(
            admin,
            collection_getter,
            admin.get_fields,
            max_number_of_rows=1
        )
        validator = admin.create_validator(model)

        class NewForm(AbstractView):

            def __init__(self, parent):
                AbstractView.__init__(self, parent)
                self.widget_layout = QtGui.QVBoxLayout()
                self.widget_layout.setMargin(0)
                title = _('new')
                index = 0
                self.form_view = admin.create_form_view(
                    title, model, index, parent
                )
                self.widget_layout.insertWidget(0, self.form_view)
                self.setLayout(self.widget_layout)
                self.validate_before_close = True
                self.entity_created_signal = SIGNAL('entity_created')
                #
                # every time data has been changed, it could become valid,
                # when this is the case, it should be propagated
                #
                self.connect(
                    model,
                    SIGNAL(
                        'dataChanged(const QModelIndex &, const QModelIndex &)'
                    ),
                    self.dataChanged
                )
                self.connect(
                    self.form_view,
                    AbstractView.title_changed_signal,
                    self.change_title
                )

            def emit_if_valid(self, valid):
                if valid:

                    def create_instance_getter(new_object):
                        return lambda:new_object[0]

                    self.emit(
                        self.entity_created_signal,
                        create_instance_getter(new_object)
                    )

            def dataChanged(self, index1, index2):

                def validate():
                    return validator.isValid(0)

                post(validate, self.emit_if_valid)

            def showMessage(self, valid):
                from camelot.view.workspace import get_workspace
                if not valid:
                    row = 0
                    reply = validator.validityDialog(row, self).exec_()
                    if reply == QtGui.QMessageBox.Discard:
                        # clear mapping to prevent data being written again to
                        # the model, after we reverted the row
                        self.form_view.widget_mapper.clearMapping()

                        def onexpunge_on_all():
                            if onexpunge:
                                for o in new_object:
                                    onexpunge(o)

                        post(onexpunge_on_all)
                        self.validate_before_close = False

                        for window in get_workspace().subWindowList():
                            if window.widget() == self:
                                window.close()
                else:
                    def create_instance_getter(new_object):
                        return lambda:new_object[0]

                    for _o in new_object:
                        self.emit(
                            self.entity_created_signal,
                            create_instance_getter(new_object)
                        )
                    self.validate_before_close = False
                    from camelot.view.workspace import NoDesktopWorkspace
                    workspace = get_workspace()
                    if isinstance(workspace, (NoDesktopWorkspace,)):
                        self.close()
                    else:
                        for window in get_workspace().subWindowList():
                            if window.widget() == self:
                                window.close()

            def validateClose(self):
                logger.debug(
                    'validate before close : %s' %
                    self.validate_before_close
                )
                if self.validate_before_close:
                    self.form_view.widget_mapper.submit()
                    logger.debug(
                        'unflushed rows : %s' %
                        str(model.hasUnflushedRows())
                    )
                    if model.hasUnflushedRows():
                        def validate(): return validator.isValid(0)
                        post(validate, self.showMessage)
                        return False
                    else:
                        return True
                return True

            def closeEvent(self, event):
                if self.validateClose():
                    event.accept()
                else:
                    event.ignore()

        form = NewForm(parent)
        if hasattr(admin, 'form_size'):
            form.setMinimumSize(admin.form_size[0], admin.form_size[1])
        return form

    @gui_function
    def create_select_view(admin, query, search_text=None, parent=None):
        """Returns a Qt widget that can be used to select an element from a
        query

        :param query: sqlalchemy query object

        :param parent: the widget that will contain this select view, the
        returned widget has an entity_selected_signal signal that will be fired
        when a entity has been selected.
        """
        from controls.tableview import TableView
        from art import Icon
        from proxy.queryproxy import QueryTableProxy
        from PyQt4.QtCore import SIGNAL

        class SelectQueryTableProxy(QueryTableProxy):
            header_icon = Icon(
                'tango/16x16/emblems/emblem-symbolic-link.png'
            ).getQIcon()

        class SelectView(TableView):

            query_table_proxy = SelectQueryTableProxy
            title_format = 'Select %s'

            def __init__(self, admin, parent):
                TableView.__init__(
                    self, admin,
                    search_text=search_text, parent=parent
                )
                self.entity_selected_signal = SIGNAL('entity_selected')
                self.connect(self, SIGNAL('row_selected'), self.sectionClicked)
                self.setUpdatesEnabled(True)

            def emit_and_close(self, instance_getter):
                self.emit(self.entity_selected_signal, instance_getter)
                from camelot.view.workspace import get_workspace
                for window in get_workspace().subWindowList():
                    if hasattr(window, 'widget') and window.widget() == self:
                        window.close()

            def sectionClicked(self, index):
                # table model will be set by the model thread, we can't
                # decently select if it has not been set yet
                if self._table_model:

                    def create_constant_getter(cst):
                        return lambda:cst

                    def create_instance_getter():
                        entity = self._table_model._get_object(index)
                        return create_constant_getter(entity)

                    post(create_instance_getter, self.emit_and_close)

        widget = SelectView(admin, parent)
        widget.setUpdatesEnabled(True)
        widget.setMinimumSize(admin.list_size[0], admin.list_size[1])
        widget.update()
        return widget

    @gui_function
    def create_table_view(self, query_getter=None, parent=None):
        """Returns a Qt widget containing a table view, for a certain query,
        using this Admin class; the table widget contains a model
        QueryTableModel

        :param query_getter: sqlalchemy query object

        :param parent: the workspace widget that will contain the table view
        """

        from PyQt4 import QtCore
        from PyQt4.QtCore import SIGNAL
        from controls.tableview import TableView
        from proxy.queryproxy import QueryTableProxy
        tableview = TableView(self)

        def createOpenForm(self, tableview):

            def openForm(index):
                from workspace import get_workspace
                model = QueryTableProxy(
                    tableview.admin,
                    tableview._table_model._query_getter,
                    tableview.admin.get_fields,
                    max_number_of_rows=1
                )
                title = ''
                formview = tableview.admin.create_form_view(
                    title, model, index, parent
                )
                get_workspace().addSubWindow(formview)
                formview.show()

            return openForm

        tableview.connect(
            tableview,
            SIGNAL('row_selected'),
            createOpenForm(self, tableview)
        )

        return tableview
    
    @model_function
    def delete(self, entity_instance):
        """Delete an entity instance"""
        from sqlalchemy.orm.session import Session
        if hasattr(entity_instance, 'id') and entity_instance.id:
                pk = entity_instance.id
                # save the state before the update
                from camelot.model.memento import BeforeDelete
                from camelot.model.authentication import getCurrentAuthentication
                history = BeforeDelete( model = unicode( self.entity.__name__ ),
                                       primary_key = pk,
                                       previous_attributes = {},
                                       authentication = getCurrentAuthentication() )
                entity_instance.delete()
                Session.object_session( entity_instance ).flush( [entity_instance] )
                Session.object_session( history ).flush( [history] )

    @model_function
    def flush(self, entity_instance):
        """Flush the pending changes of this entity instance to the backend"""
        from sqlalchemy.orm.session import Session
        Session.object_session( entity_instance ).flush( [entity_instance] )



from PyQt4 import QtGui
from PyQt4.QtCore import Qt

from camelot.view.controls import editors
from camelot.core.utils import variant_to_pyobject

import logging
logger = logging.getLogger( 'camelot.view.controls.delegates.one2manydelegate' )

class One2ManyDelegate( QtGui.QItemDelegate ):
    """Custom delegate for many 2 one relations
  
  .. image:: ../_static/onetomany.png  
  """

    def __init__( self, parent = None, **kwargs ):
        logger.debug( 'create one2manycolumn delegate' )
        assert 'delegate' in kwargs
        QtGui.QItemDelegate.__init__( self, parent )
        self.kwargs = kwargs

    def createEditor( self, parent, option, index ):
        logger.debug( 'create a one2many editor' )
        editor = editors.One2ManyEditor( parent = parent, **self.kwargs )
        self.setEditorData( editor, index )
        return editor

    def setEditorData( self, editor, index ):
        logger.debug( 'set one2many editor data' )
        model = variant_to_pyobject( index.data( Qt.EditRole ) )
        editor.set_value( model )

    def setModelData( self, editor, model, index ):
        pass



from filedelegate import FileDelegate
from camelot.view.controls import editors

class ImageDelegate(FileDelegate):
    """
    .. image:: ../_static/image.png
    """
    
    editor = editors.ImageEditor




"""
Form actions are objects that can be put in the form_actions list of the
object admin interfaces.  Those actions then appear on the form and can
be executed by the end users.  The most convenient method to create custom actions
is to subclass FormAction and implement a custom run method ::

  class MyAction(FormAction):

    def run(self, entity_getter):
      print 'Hello World'

  class Movie(Entity):
    title = Field(Unicode(60), required=True)

    Admin(EntityAdmin):
      list_display = ['title']
      form_actions = [MyAction('Hello')]

Several subclasses of FormAction exist to provide common use cases such as executing
a function in the model thread or printing a report.

To customize the look of the action button on the form, the render method should be
overwritten.
"""

from PyQt4 import QtGui, QtCore
from camelot.view.art import Icon
from camelot.view.model_thread import gui_function, post

class FormAction( object ):
    """Abstract base class to implement form actions"""

    def __init__( self, name, icon = None ):
        self._name = name
        self._icon = icon

    @gui_function
    def render( self, parent, entity_getter ):
        """Returns a QWidget the user can use to trigger the action"""

        def create_clicked_function( self, entity_getter ):

            def clicked( *args ):
                self.run( entity_getter )

            return clicked

        button = QtGui.QPushButton( unicode(self._name) )
        if self._icon:
            button.setIcon( self._icon.getQIcon() )
        button.connect( button, QtCore.SIGNAL( 'clicked()' ), create_clicked_function( self, entity_getter ) )
        return button

    @gui_function
    def run( self, entity_getter ):
        """Overwrite this method to create an action that does something"""
        raise NotImplementedError

class FormActionFromGuiFunction( FormAction ):
    """Convert a function that is supposed to run in the GUI thread to a FormAction"""

    def __init__( self, name, gui_function, icon = None ):
        FormAction.__init__( self, name, icon )
        self._gui_function = gui_function

    @gui_function
    def run( self, entity_getter ):
        self._gui_function( entity_getter )

class FormActionProgressDialog(QtGui.QProgressDialog):

    def __init__(self, name):
        QtGui.QProgressDialog.__init__( self, 'Please wait', QtCore.QString(), 0, 0 )
        self.setWindowTitle( unicode(name) )

    def finished(self, success):
        self.close()

    def print_result(self, html):
        from camelot.view.export.printer import open_html_in_print_preview_from_gui_thread
        self.close()
        open_html_in_print_preview_from_gui_thread(html)


class FormActionFromModelFunction( FormAction ):
    """Convert a function that is supposed to run in the model thread to a FormAction"""

    def __init__( self, name, model_function, icon = None ):
        FormAction.__init__( self, name, icon )
        self._model_function = model_function

    @gui_function
    def run( self, entity_getter ):
        progress = FormActionProgressDialog(self._name)

        def create_request( entity_getter ):

            def request():
                o = entity_getter()
                self._model_function( o )
                return True

            return request

        post( create_request( entity_getter ), progress.finished, exception = progress.finished )
        progress.exec_()

class PrintHtmlFormAction( FormActionFromModelFunction ):
    """Create an action for a form that pops up a print preview for generated html.
  Overwrite the html function to customize the html that should be shown::

    class PrintMovieAction(PrintHtmlFormAction):

      def html(self, movie):
        html = '<h1>' + movie.title + '</h1>'
        html += movie.description
      return html

    class Movie(Entity):
      title = Field(Unicode(60), required=True)
      description = Field(camelot.types.RichText)

      class Admin(EntityAdmin):
        list_display = ['title', 'description']
        form_actions = [PrintMovieAction('summary')]

  will put a print button on the form :

  .. image:: ../_static/formaction/print_html_form_action.png

    """

    def __init__( self, name, icon = Icon( 'tango/16x16/actions/document-print.png' ) ):
        FormActionFromModelFunction.__init__( self, name, self.html, icon )

    def html( self, o ):
        """Overwrite this function to generate custom html to be printed
        :arg o: the object that is displayed in the form"""
        return '<h1>' + unicode( o ) + '<h1>'

    @gui_function
    def run( self, entity_getter ):
        progress = FormActionProgressDialog(self._name)

        def create_request( entity_getter ):

            def request():
                o = entity_getter()
                return self.html( o )

            return request

        post( create_request( entity_getter ), progress.print_result, exception = progress.finished )
        progress.exec_()

def structure_to_form_actions( structure ):
    """Convert a list of python objects to a list of form actions.  If the python
    object is a tuple, a FormActionFromGuiFunction is constructed with this tuple as arguments.  If
    the python object is an instance of a FormAction, it is kept as is.
    """

    def object_to_action( o ):
        if isinstance( o, FormAction ):
            return o
        return FormActionFromGuiFunction( o[0], o[1] )

    return [object_to_action( o ) for o in structure]



from fileeditor import FileEditor, CustomEditor
from wideeditor import WideEditor
from camelot.view.art import Icon

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt
from camelot.core.utils import ugettext_lazy as _

class ImageEditor(FileEditor, WideEditor):
    """Editor to view and edit image files, this is a customized implementation
    of a FileEditor"""
    
    filter = """Image files (*.bmp *.jpg *.jpeg *.mng *.png *.pbm *.pgm *.ppm *.tiff *.xbm *.xpm)
All files (*)"""
    
    def __init__(self, 
                 parent=None, 
                 editable=True, 
                 storage=None,
                 preview_width=100,
                 preview_height=100, 
                 **kwargs):
        self.preview_width = preview_width
        self.preview_height = preview_height
        FileEditor.__init__(self, parent=parent, storage=storage, editable=editable, **kwargs)
      
    def setup_widget(self):
        self.layout = QtGui.QHBoxLayout()
        self.layout.setContentsMargins( 0, 0, 0, 0 )
        #
        # Setup label
        #
        self.label = QtGui.QLabel(self)
        self.label.setEnabled(self.editable)
        self.layout.addWidget(self.label)
        self.label.setAlignment(Qt.AlignHCenter|Qt.AlignVCenter)
        #
        # Setup buttons
        #
        button_layout = QtGui.QVBoxLayout()
        button_layout.setSpacing(0)
        button_layout.setMargin(0)
        
        self.open_button = QtGui.QToolButton()
        self.open_button.setIcon(self.open_icon)
        self.open_button.setEnabled(self.editable)
        self.open_button.setAutoRaise(True)
        self.open_button.setToolTip(unicode(_('open image')))
        self.connect(self.open_button, QtCore.SIGNAL('clicked()'), self.open_button_clicked)
        
        self.clear_button = QtGui.QToolButton()
        self.clear_button.setIcon(self.clear_icon)
        self.clear_button.setEnabled(self.editable)
        self.clear_button.setToolTip(unicode(_('delete image')))
        self.clear_button.setAutoRaise(True)
        self.connect(self.clear_button, QtCore.SIGNAL('clicked()'), self.clear_button_clicked)

        button_layout.addStretch() 
        button_layout.addWidget(self.open_button)
        button_layout.addWidget(self.clear_button)    
        
        self.layout.addLayout(button_layout)
        self.layout.addStretch()                
        self.setLayout(self.layout)
        self.clear_image()
      
    def set_enabled(self, editable=True):
        self.clear_button.setEnabled(editable)
        self.open_button.setEnabled(editable)
              
    def set_pixmap(self, pixmap):
        self.label.setPixmap(pixmap)      
        self.draw_border()
        
    def set_image(self, image):
        self.set_pixmap(QtGui.QPixmap.fromImage(image))
    
    def clear_image(self):
        dummy_image = Icon('tango/32x32/mimetypes/image-x-generic.png')
        self.set_pixmap(dummy_image.getQPixmap())
              
    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        if value:
            self.open_button.setIcon(self.open_icon)
            self.open_button.setToolTip(unicode(_('open file')))
            if value!=self.value:
                from camelot.view.model_thread import post
                post(lambda:value.checkout_thumbnail(self.preview_width,self.preview_height), self.set_image)            
        else:
            self.clear_image()
            self.open_button.setIcon(self.new_icon)
            self.open_button.setToolTip(unicode(_('add file')))
        self.value = value
        return value
    
    def draw_border(self):
        self.label.setFrameShape(QtGui.QFrame.Box)
        self.label.setFrameShadow(QtGui.QFrame.Plain)
        self.label.setLineWidth(1)
        self.label.setFixedSize(self.preview_width, self.preview_height)



from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.core.utils import variant_to_pyobject
from camelot.core import constants
from camelot.view.proxy import ValueLoading

class CurrencyDelegate(CustomDelegate):
    """Custom delegate for float values"""
  
    __metaclass__ = DocumentationMetaclass
    
    editor = editors.FloatEditor
    
    def __init__(self,
                 minimum=constants.camelot_minfloat,
                 maximum=constants.camelot_maxfloat,
                 precision=2,
                 editable=True,
                 parent=None,
                 prefix="",
                 suffix="",
                 **kwargs):
        """
    :param precision:  The number of digits after the decimal point displayed.  This defaults
    to the precision specified in the definition of the Field.     
    """
        CustomDelegate.__init__(self, parent=parent, editable=editable, minimum=minimum, maximum=maximum,
                                precision=precision, prefix=prefix, suffix=suffix, **kwargs)
        self.minimum = minimum
        self.maximum = maximum
        self.precision = precision
        self.editable = editable
        self.prefix = prefix
        self.suffix = suffix
        
    def setEditorData(self, editor, index):
        value = variant_to_pyobject(index.model().data(index, Qt.EditRole))
        editor.set_value(value)
    
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        value = variant_to_pyobject(index.model().data(index, Qt.EditRole))
        if value in (None, ValueLoading):
            value = 0.0
        rect = option.rect
        rect = QtCore.QRect(rect.left()+3, rect.top()+6, 16, 16)
        
        
        background_color = QtGui.QColor(index.model().data(index, Qt.BackgroundRole))
        
        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect(option.rect, option.palette.highlight())
            fontColor = QtGui.QColor()
            if self.editable:
                Color = option.palette.highlightedText().color()
                fontColor.setRgb(Color.red(), Color.green(), Color.blue())
            else:
                fontColor.setRgb(130,130,130)
        else:
            if self.editable:
                painter.fillRect(option.rect, background_color)
                fontColor = QtGui.QColor()
                fontColor.setRgb(0,0,0)
            else:
                painter.fillRect(option.rect, option.palette.window())
                fontColor = QtGui.QColor()
                fontColor.setRgb(130,130,130)
                
        value_str_formatted  = QtCore.QString("%L1").arg(value,0,'f',2)
        
        painter.setPen(fontColor.toRgb())
        rect = QtCore.QRect(option.rect.left()+23,
                            option.rect.top(),
                            option.rect.width()-23,
                            option.rect.height())
        
        painter.drawText(rect.x()+2,
                         rect.y(),
                         rect.width()-4,
                         rect.height(),
                         Qt.AlignVCenter | Qt.AlignRight,
                         value_str_formatted)
        
        painter.restore()
